

'''
Fabric functionality for configuring, deploying, and starting python 
applications.

This module provides an API to access directories where an application stores
configuration files, log files, a virtual environment, a binaries, and the
webapp code.

This module encodes a very arbitrary and specific way of organizing an
application and so might not be useful to a wide audience.

'''

import os


class App(object):
    '''
    PROJ DIRECTORY STRUCTURE

    root/
        bin/ # executables
        app/ # app code
        log/ # log files
        conf/ # configuration
        venv/ # virtual environment

    YMMV with this structure. Maybe you want your log files under /var/log/APP
    or you have a different name for your app dir.  The point is to create a
    API for getting the locations without hardcoding the structure in dependent
    code, so the structure can vary without breaking the dependent code.
    '''

    def __init__(self, root):
        '''
        root: the root directory of the application project
        '''
        self.root = root

    def bindir(self):
        return os.path.join(self.root, 'bin')

    def appdir(self):
        return os.path.join(self.root, 'app')

    def logdir(self):
        return os.path.join(self.root, 'log')

    def confdir(self):
        return os.path.join(self.root, 'conf')

    def venvdir(self):
        '''
        returns: the path to the virtual environment
        '''
        return os.path.join(self.root, 'venv')







'''
FABRIC UTILITY FUNCTIONS
'''


import collections
import os
import shutil
import subprocess

import fabric.api
import fabric.contrib.files
import fabric.operations
from fabric.api import env, task, sudo, run, cd, local, lcd, execute, get, put
from fabric.contrib.files import exists, upload_template
from fabric.contrib.project import rsync_project


###############
# EC2 FUNCTIONS

def print_instances(instances, prefix=''):
    for instance in instances:
        print_instance(instance, prefix=prefix)


def print_instance(instance, prefix=''):
    print '{}Instance id={}, state={}, tags={}, public_dns_name={}'.format(
        prefix, instance.id, instance.state, instance.tags,
        instance.public_dns_name)


def terminate_instances(conn, instances):
    if not instances:
        return
    killed_instances = conn.terminate_instances([i.id for i in instances])
    if len(killed_instances) != len(instances):
        raise Exception('Not all instances terminated.', instances, 
                        killed_instances)
    print 'Terminated instances:'
    print_instances(killed_instances, '\t')


def get_named_instance(conn, name):
    '''
    Return a single non-terminated, non-shutting-down
    boto.ec2.instance.Instance that has a tag 'Name' that
    matches ``name``.

    conn: a boto.ec2.connection.EC2Connection
    name: string. The unique value of the 'Name' tag of a (running) instance
    return: the instance of the current webserver, of None if there is none.
    raise: Exception if there is more than one non-terminated instance.
    '''
    rs = conn.get_all_instances(filters={'tag:Name': name})
    instances = [i for r in rs for i in r.instances if i.state not in ['terminated', 'shutting-down']]
    if len(instances) > 1:
        raise Exception('More than one instance', instances)
    if not instances:
        return None
    return instances[0]





##########################
# PROJ DIRECTORY STRUCTURE

# dir/
#   bin/ # executables
#   app/ # app
#   log/ # log files
#   conf/ # configuration
#   venv/ # virtual environment

# YMMV with this structure. Maybe you want your log files under /var/log/APP
# or you have a different name for your app dir.
# The point of dir_config is to create a clean API for getting the
# locations without hardcoding the structure in dependent code,
# so it can vary without breaking the dependent code.

def bin_dir(dir):
    return os.path.join(dir, 'bin')


def app_dir(dir):
    return os.path.join(dir, 'app')


def log_dir(dir):
    return os.path.join(dir, 'log')


def conf_dir(dir):
    return os.path.join(dir, 'conf')


def venv_dir(dir):
    '''
    dir: parent directory above the virtual environment dir
    returns: the path to the virtual environment
    '''
    return os.path.join(dir, 'venv')

def dir_config(dir, config):
    config.venv = venv_dir(dir)
    config.log = log_dir(dir)
    config.conf = conf_dir(dir)
    config.bin = bin_dir(dir)
    config.app = app_dir(dir)


######################

#############
# SUPERVISORD

def config_supervisord_program(src, dest, context=None, use_sudo=False,
                               mode=None):
    '''
    Upload supervisor config for a program and reload/restart.
    '''
    upload_template(
        os.path.join(conf_dir(HERE), 'heyluca_supervisord.conf'),
        '/etc/supervisor.d/heyluca.conf',
        context={'name': 'heyluca',
                 'venv': conf().venv,
                 'app': conf().app,
                 'entry': 'heyluca:app',
                 'conf': conf().conf,
                 'log': conf().log,
                 'user': conf().user,
                },
        use_sudo=True,
        mode=0770)




    

############
# DEPRECATED

# I now recommend running sshd on localhost to maintain the Fabric paradigm.

# This is bad advice:
# http://stackoverflow.com/questions/6725244/running-fabric-script-locally
def setremote():
    '''
    set env.* to the remote versions for run, cd, rsync, and exists
    '''
    env.run = fabric.api.run
    env.cd = fabric.api.cd
    env.rsync = rsync
    env.exists = fabric.contrib.files.exists
    return env

def setlocal():
    '''
    set env.* to the localhost versions for run, cd, rsync, and exists
    '''
    env.run = fabric.api.local
    env.cd = fabric.api.lcd
    env.rsync = lrsync
    env.exists = os.path.exists
    return env


def lrsync(options, src, dest, cwd=None, **kwds):
    '''
    options: list of rsync options, e.g. ['--delete', '-avz']
    src: source directory (or files).  Note: rsync behavior varies depending on whether or not src dir ends in '/'.
    dest: destination directory.
    cwd: change (using subprocess) to cwd before running rsync.
    This is a helper function for using rsync to sync files to localhost.
    It uses subprocess.  Note: shell=False.
    Use rsync() to sync files to a remote host.
    '''
    args = ['rsync'] + options + [src, dest]
    print args
    subprocess.check_call(args, cwd=cwd)
    
    






'''
Fabric-based functions working with python virtual environments:

- creating a new virtual environment
- remving an existing virtual environment
- installing requirements in a virtual environment
- "freezing" requirements from a virtual environment
- getting paths of executables, etc., within a virtual environment.
'''



import os

from fabric.api import run, put, get
from fabric.contrib.files import exists


def bin(venv):
    '''
    return: path to the venv bin dir.
    '''
    return os.path.join(venv, 'bin')


def python(venv):
    '''
    return: path to the venv python executable.
    '''
    return os.path.join(venv, 'bin', 'python')


def pip(venv):
    '''
    return: path to the venv pip executable.
    '''
    return os.path.join(venv, 'bin', 'pip')


def remove(venv):
    '''
    Remove the virtual environment completely
    '''
    if exists(venv):
        print 'cleaning', venv
        run('rm -rf {}'.format(venv))


def create(venv, python='python', virtualenv_script=None):
    '''
    venv: virtual environment directory to create.  venv MUST NOT already exist.
    python: path or name of python executable to use to create the venv.
    Defaults to 'python'.
    virtualenv_script: Optional local path to virtualenv.py.  If given, it will
    be copied to venv and run by python to create venv.  If not given,
    virtualenv.py will be downloaded from the internet, which could be less
    reliable, slower, and does not guarantee a fixed version.  

    Create a virtual environment located at venv on the remote host.  Raise an exception if venv already exists.
    '''
    script_path = os.path.join(venv, 'virtualenv.py')
    script_url = 'https://raw.github.com/pypa/virtualenv/master/virtualenv.py'

    # require that the venv does not yet exist.
    if exists(venv):
        raise Exception('Path already exists. Abort creation. venv={}'.format(venv))

    # create the venv dir for virtualenv.py
    run('mkdir -p {}'.format(venv))

    # put virtualenv.py in venv or download it.
    if virtualenv_script:
        put(virtualenv_script, script_path)
    else:
        run('curl -o {} {}'.format(script_path, script_url))

    # create the venv
    run('{} {} --distribute {}'.format(python, script_path, venv))


def install(venv, requirements):
    '''
    venv: virtual environment directory to create.
    requirements: local path of requirements.txt file to be copied to venv dir
    to the virtual environment and used to install packages.
    Use the venv pip to install the requirements file.
    '''
    remote_path = os.path.join(venv, 'requirements.txt')
    put(requirements, remote_path)
    run('{} install -r {}'.format(pip(venv), remote_path))



def freeze(venv, requirements):
    '''
    venv: virtual environment directory to freeze.
    requirements: local path in which to save output of pip freeze.
    '''
    remote_path = os.path.join(venv, 'requirements.txt')
    run('{} freeze > {}'.format(pip(venv), remote_path))
    get(remote_path, requirements)







'''
Fabric utilities for working with files.
'''

import StringIO
import os
import subprocess

from fabric.api import sudo, run, settings, hide, put
from fabric.contrib.files import exists



################
# FILE FUNCTIONS


def fix_shebang(shebang, handle):
    '''
    shebang: a shebang line, e.g. #!/usr/bin/env python or #!/bin/sh.  If
    shebang does not start with '#!', then '#!' will be prepended to it.  If
    shebang does not end with a newline, a newline will be appended.

    handle: a iterable of lines, presumably the contents of a file that needs a
    shebang line or a new shebang line.

    Yield shebang and then the lines in handle except the first line in handle
    if it is a shebang line.
    '''
    # make sure shebang is starts with '#!' and ends with a newline.
    if not shebang.startswith('#!'):
        shebang = '#!' + shebang

    if not shebang.endswith('\n'):
        shebang += '\n'

    for i, line in enumerate(handle):
        if i == 0:
            yield shebang
            if not line.startswith('#!'):
                yield line
        else:
            yield line


def upload_shebang(filename, destination, shebang, use_sudo=False, backup=True,
                   mirror_local_mode=False, mode=None):
    """
    Upload a text file to a remote host, adding or updating the shebang line.

    ``filename`` should be the path to a text file.

    ``shebang`` should be a string containing a shebang line.  E.g.
    "#!/usr/bin/python\n".  If shebang does not start with '#!' or end with a
    newline, these will be added.

    If the first line in filename starts with '#!' it will be replaced with
    shebang.  If the first line does not start with #!, shebang will be
    prepended to the contents of filename.

    The resulting file will be uploaded to the remote file path
    ``destination``.  If the destination file already exists, it will be
    renamed with a ``.bak`` extension unless ``backup=False`` is specified.

    By default, the file will be copied to ``destination`` as the logged-in
    user; specify ``use_sudo=True`` to use `sudo` instead.

    The ``mirror_local_mode`` and ``mode`` kwargs are passed directly to an
    internal `~fabric.operations.put` call; please see its documentation for
    details on these two options.
    """
    func = use_sudo and sudo or run
    # Normalize destination to be an actual filename, due to using StringIO
    with settings(hide('everything'), warn_only=True):
        if func('test -d %s' % destination).succeeded:
            sep = "" if destination.endswith('/') else "/"
            destination += sep + os.path.basename(filename)

    # Use mode kwarg to implement mirror_local_mode, again due to using
    # StringIO
    if mirror_local_mode and mode is None:
        mode = os.stat(filename).st_mode
        # To prevent put() from trying to do this
        # logic itself
        mirror_local_mode = False

    # process filename
    text = None
    with open(filename) as inputfile:
        text = ''.join(fix_shebang(shebang, inputfile))

    # Back up original file
    if backup and exists(destination):
        func("cp %s{,.bak}" % destination)

    # Upload the file.
    put(
        local_path=StringIO.StringIO(text),
        remote_path=destination,
        use_sudo=use_sudo,
        mirror_local_mode=mirror_local_mode,
        mode=mode
    )


def upload_format(filename, destination, args=None, kws=None,
                  use_sudo=False, backup=True, mirror_local_mode=False,
                  mode=None):
    """
    Read in the contents of filename, format the contents via
    contents.format(*args, **kws), and upload the results to the
    destination on the remote host.

    ``filename`` should be the path to a text file.  The contents of
    ``filename`` will be read in.

    Format the contents, using contents.format(*args, **kws).  If
    args is None, it will not be included in the format() call.
    Likewise for kws.

    The resulting contents will be uploaded to the remote file path
    ``destination``.  If the destination file already exists, it will be
    renamed with a ``.bak`` extension unless ``backup=False`` is specified.

    By default, the file will be copied to ``destination`` as the logged-in
    user; specify ``use_sudo=True`` to use `sudo` instead.

    The ``mirror_local_mode`` and ``mode`` kwargs are passed directly to an
    internal `~fabric.operations.put` call; please see its documentation for
    details on these two options.
    """
    func = use_sudo and sudo or run
    # Normalize destination to be an actual filename, due to using StringIO
    with settings(hide('everything'), warn_only=True):
        if func('test -d %s' % destination).succeeded:
            sep = "" if destination.endswith('/') else "/"
            destination += sep + os.path.basename(filename)

    # Use mode kwarg to implement mirror_local_mode, again due to using
    # StringIO
    if mirror_local_mode and mode is None:
        mode = os.stat(filename).st_mode
        # To prevent put() from trying to do this
        # logic itself
        mirror_local_mode = False

    # process filename
    text = None
    with open(filename) as inputfile:
        if not args:
            args = []
        
        if not kws:
            kws = {}

        text = inputfile.read().format(*args, **kws)

    # Back up original file
    if backup and exists(destination):
        func("cp %s{,.bak}" % destination)

    # Upload the file.
    put(
        local_path=StringIO.StringIO(text),
        remote_path=destination,
        use_sudo=use_sudo,
        mirror_local_mode=mirror_local_mode,
        mode=mode
    )


def file_format(infile, outfile, args=None, kws=None):
    '''
    Consider using fabric.contrib.files.upload_template or upload_format
    infile: a local file path
    outfile: a local file path.
    Read the contents of infile as a string, ''.format() the string using args
    and kws, and write the formatted string to outfile.  This is useful if
    infile is a "template" and args and kws contain the concrete values
    for the template.
    '''
    if args is None:
        args = []
    if kws is None:
        kws is {}
    with open(infile) as fh:
        text = fh.read()
    new_text = text.format(*args, **kws)
    with open(outfile, 'w') as fh2:
        fh2.write(new_text)


def rsync(options, src, dest, user=None, host=None, cwd=None):
    '''
    Consider using fabric.contrib.project.rsync_project.
    options: list of rsync options, e.g. ['--delete', '-avz']
    src: source directory (or files).  Note: rsync behavior varies depending on whether or not src dir ends in '/'.
    dest: destination directory.
    cwd: change (using subprocess) to cwd before running rsync.
    This is a helper function for running rsync locally, via subprocess.  Note: shell=False.
    '''
    # if remote user and host specified, copy there instead of locally.
    if user and host:
        destStr = '{}@{}:{}'.format(user, host, dest)
    else:
        destStr = dest

    args = ['rsync'] + options + [src, destStr]
    print args
    subprocess.check_call(args, cwd=cwd)





from setuptools import setup

setup(
    name = 'diabric',
    version = '0.1',
    license = 'MIT',
    description = 'Diabolically atomic Python Fabric fabfile tasks and utilities.',
    long_description = open('README.md').read(),
    keywords = 'fabric fabfile boto ec2 virtualenv python wsgi webapp deployment',
    download_url = 'https://github.com/todddeluca/diabric/downloads',
    url = 'https://github.com/todddeluca/diabric',
    author = 'Todd Francis DeLuca',
    author_email = 'todddeluca@yahoo.com',
    classifiers = ['License :: OSI Approved :: MIT License',
                   'Development Status :: 2 - Pre-Alpha',
                   'Programming Language :: Python :: 2',
                   'Programming Language :: Python :: 2.7',
                  ],
    packages = ['diabric'],
    install_requires = ['setuptools', 'Fabric>=1.4','boto>=2.3'],
    include_package_data = True,
    package_data = {'' : ['README.md', 'LICENSE.txt']},
)





'''

With this module you can:

- Create configurations for different contexts.
- Access configurations depending on the context of a task.

Best practices for Fabric configuration:

- Avoid storing configuration in fabric.api.env.  Leave env for Fabric-specific
  configuration.  An example of what happens when you do not.  If you want to
  configure the port a server will listen on, you might store it in env.port,
  overwriting the port fabric uses for ssh.  Chaos and failure ensues.
'''


import collections

from fabric.api import env


class AttrDict(dict):
    '''
    A dictionary whose keys can also be accessed as items or attributes.
    This provides syntactic sugar to avoid typing lots of brackets and quotes.
    However by conflating attribute and item access semantics it can lead to
    strange behavior.

    Example:

        obj = AttrDict()
        obj['key1'] = 'hi'
        obj.key2 = 'hello'
        print obj['key1']
        print obj.key2

    Example of strange behavior when setting a built-in attribute:

        obj = AttrDict()
        obj.hi = 'hello!'
        print obj.get('hi') # 'hello!'
        obj.get = 'bye' # overwrite dict get function?  No.
        print obj.get # <built-in method get of AttrDict object at ...>
        print obj['get'] # 'bye'
        print obj.get('hi') # 'hello!'.  get() still works.
    '''
    def __getattr__(self, key):
        return self[key]

    def __setattr__(self, key, value):
        self[key] = value


class AttrDictContextConfig(collections.defaultdict):
    '''
    AttrDictContextConfig is a defaultdict of AttrDict objects used to store configuration.
    It is also a callable that returns the AttrDict for the current context.

    Usage example:

        # Create a config collection for each role
        config = AttrDictContextConfig(role_context)

        # Set configuration values for each role.
        config['dev'].deploy_dir = '/www/example.com'
        config['dev'].user = 'dev-user'
        config['prod'].deploy_dir = '/www/dev.example.com'
        config['prod'].user = 'prod-user'

        # Use the configuration in a task
        @task
        def deploy():
            put('a_file.txt', config().deploy_dir)

        # Invoke the task with a role
        fab -R prod deploy
    '''

    def __init__(self, context):
        '''
        context: a function that returns a key for whatever the current
        configuration context is.  For example, host_context returns the
        current env.host of a task, which would be useful if configuration
        varied by host.
        '''
        super(AttrDictContextConfig, self).__init__(AttrDict)
        self.context = context

    def __call__(self):
        '''
        Choose a configuration depending on the current context key.
        '''
        return self[self.context()]


class ContextConfig(collections.defaultdict):
    '''
    ContextConfig is a defaultdict of dict objects used to store configuration.
    It is also a callable that returns the dict for the current context.

    Usage example:

        # Create a config collection for each role
        config = ContextConfig(role_context)

        # Set configuration values for each role.
        config['dev']['deploy_dir'] = '/www/example.com'
        config['dev']['user'] = 'dev-user'
        config['prod']['deploy_dir'] = '/www/dev.example.com'
        config['prod']['user'] = 'prod-user'

        # Use the configuration in a task
        @task
        def deploy():
            put('a_file.txt', config()['deploy_dir']

        # Invoke the task with a role
        fab -R prod deploy
    '''

    def __init__(self, context):
        '''
        context: a function that returns a key for whatever the current
        configuration context is.  For example, host_context returns the
        current env.host of a task, which would be useful if configuration
        varied by host.
        '''
        super(ContextConfig, self).__init__(dict)
        self.context = context

    def __call__(self):
        '''
        Choose a configuration depending on the current context key.
        '''
        return self[self.context()]


def host_context():
    '''
    Returns the current fabric env.host.  This is useful with
    KeyConfig(host_context) to get the configuration associated with the
    current host in a task.
    '''
    return env.host


def role_context():
    '''
    Return the first role in the current fabric env.roles list, or None if
    there are no roles.  This is useful with KeyConfig(role_context) to get the
    configuration for the current role in a task.
    '''
    if len(env.roles):
        return env.roles[0]
    else:
        return None


# Example of using AttrDict for configuration

# Create a collection of configs for each environment
# config = collections.defaultdict(AttrDict)

# Set configuration values for each environment
# config['dev'].system_python = '/usr/bin/python2.7'
# config['dev'].deploy_dir = '/www/example.com'
# config['dev'].user = 'dev-user'

# config['prod'].system_python = '/usr/bin/python2.7'
# config['prod'].deploy_dir = '/www/dev.example.com'
# config['prod'].user = 'prod-user'

# Set derived config values
# for c in config.values():
#     c.venv = venv_dir(c.deploy_dir)
#     c.log = log_dir(c.deploy_dir)
#     c.conf = conf_dir(c.deploy_dir)
#     c.bin = bin_dir(c.deploy_dir)
#     c.app = app_dir(c.deploy_dir)

# Create a convenience function to retrieve the configuration dict
# for the current environment. Use conf().key instead of env.key.
# def conf():
#     '''
#     Choose a configuration depending on the current role.
#     '''
#     return config[env.roles[0]]



